;*************************************************************************
;** 	Software UART 					Rev.1.01	**
;**     8N1, single wire/half duplex (open drain: needs pullup-Resistor)**
;** 	MCU type: enhanced midrange PIC microcontroller			**
;*************************************************************************
;R1.01 Bugfix (Improper reload of timer with values > 0x10)

; Used Hardware:
; 	TIMER1 + interrupt
;	Single IO-pin for TxD/RxD with interrupt on change function.

;*********************** customization *********************
;**** put the following definitions in main source code
;**** and adjust the numbers as needed: 
  ifdef __16F1825
	#include <P16F1825.INC>         	; processor specific variable definitions
  endif

#include picalic_config.inc

#ifdef DEBUG_UART

	global uartCR, uartDR, UartTxByte, intTMR1, intIOC, init_UART
	extern MonRxByte

	radix dec
#define BAUDRATE 9600		;UART baud rate 
#define UTXBUFSIZE 8		;size of UART Tx Buffer (must be an integer power of 2!)
#define FOSC 32000000		;clock oscillator frequency, used to calculate latencies
#define TMR1CLK 2000000		;TMR1 clock frequency (output of prescaler)
#define UART_PORT PORTA		;Tx/Rx-Pin: PORTA or PORTB, 
#define UART_PIN  0		;pin must have IOC function!

;uartCR status Bits:
#define RXBSY 0		;receiving
#define TXBSY 1		;transmitting
#define RXBF 2			;receiver buffer full (rx-data in uartDR)
#define TXBF 3			;txbuffer contains data to send
#define STASTO 4		;rx: expect stopbit, tx: just sent startbit 
#define COLLISION 5		;set to 1 if a data collision was detected (only
;				valid if enabled)
;#define COLLISIONDETECT	;if defined: detection of data collision is enabled

;*********************** RAM usage ************************
;put the following variables into BANK0 section:

UARTVARS udata 0x49
uartCR res 1				;controlregister of software-UART
uartSR res 1				;shift register
uartDR res 1				;Rx data register
uartTxRd res 1				;Tx buffer pointers...
uartTxWr res 1
uartTxBuffer res UTXBUFSIZE		;... and buffer memory

;*************************************************************************************
; recommended interrupt program code:
;	org 4
;interrupt
;	banksel	IOCAF
;	movf	IOCAF,W		;change of input pin detected?
;	btfss	STATUS,Z
;	goto	intIOC
;
;	banksel	0
;	btfsc	PIR1,TMR1IF	;timer overflow of timer 1?
;	goto	intTMR1		;y: UART I/O
;	
;	... (ISR code for other interrupt sources)
;
;intIOC
;	btfsc	IOCxF,UART_PIN		;(x = A or B, according to Tx/Rx-Pin location
;	goto	IOCUartInt		;H->L transition: UART Rx start bit
;	... (ISR code for other IOC sources)
;*************************************************************************************
; Usage:
;
;Initialization:
;Upon Reset, initialize UART's Tx/Rx-Pin to digital input. Choose appropriate clock
;source and prescaler setting for Timer1, but keep TMR1 stopped (T1ON = 0).
;Enable TMR1 interrupt and pin change interrupt for Rx/Tx-pin @ negative transition.
;All UART variables have to be initialized to 0.  
;
;Transmission:
;Call UartTxByte with data byte to be transmitted in W register.
;If Tx/Rx is not busy, UartTxByte issues transmission of data immediately.
;If Tx/Rx is busy, UartTxByte writes the data into buffer (uartTxBuffer).
;All data from uartTxBuffer will be sent automatically vie ISR.
;UartTxByte does not block program execution and may even be called from within an 
;interrupt service routine.
;UartTxByte does not care about buffer overflow, therefore the user is responsible
;for having enough space in uartTxBuffer before calling UartTxByte.

;Receiption:
;After a databyte has been received successfully (i.e. without framing error), the Rx-ISR 
;copies that data into uartDR and sets the RXBF-bit in UART control register (uartCR).
;The user must read uartDR and reset the RXBF bit before the succeeding databyte's
;stopbit gets received. 

;Data collision:
;The transmitter does not start transmission while the receiver is busy.
;During transmission, the receiver is inactive. Therefore comparing Rx-data with Tx-data
;cannot be used for detection of data collisions.
;If COLLISIONDETECT is defined, the transmitter reads the level on the TxD-Pin 
;right before sending the next bit and sets COLLISION bit in uartCR if a mismatch
;with the previously sent data bit has been detected. 

;Interferences with other interrupts:
;The generation and reading of of Tx/Rx signals is controlled by timer interrupts.
;Occurance of other interrupts during Tx/Rx may lead to timing jitter of TxD signals or 
;delayed scan time for RxD. With short ISR execution times while the baudrate is low,
;generally this will not cause problems. Long running ISRs and/or higher baud rates might 
;cause data corruption. Even short ISRs might cause data corruption if very high baudrates
;are used. In such cases, it may be necessary to disable all interrupts except TMR1 and 
;UART's pin change interrupt during the data transfer.      
;******************************************************************************************

TMR1RLD set (TMR1CLK/BAUDRATE)	;TMR1 reload value to generate bit interval timing

;latency: IRQ(neg.transition of RX-startbit)->TMR1ON + TMR1-IRQ->scan Rx-level = ca.40 instr.cycles = 160 clocks.
TMR1_LATE1 set (160*TMR1CLK)/FOSC	;= correction of timer1 load value (Rx Startbit). 

#ifdef COLLISIONDETECT
;latency: (TMR1-IRQ->output 1st bit) - (TMR1ON->output startbit) = ca.36-8 = 28 cycles = 112 clocks	
TMR1_LATE2 set (112*TMR1CLK)/FOSC	;= correction of timer1 load value (Start of Tx). 
#else
;latency: (TMR1-IRQ->output 1st bit) - (TMR1ON->output startbit) = ca.30-8 = 22 cycles = 88 clocks	
TMR1_LATE2 set (88*TMR1CLK)/FOSC	;= correction of timer1 load value (Start of Tx). 
#endif

UARTCODE code
 errorlevel-302

init_UART
	banksel	IOCAF
	movlw	1
	movwf	IOCAN			;enable RA0 negative transition interrupt
	clrf	IOCAF
	banksel	PIE1
	bsf	PIE1,TMR1IE		;enable TMR1 Interrupt
	banksel	0
	movlw	b'00100000'		;T1 counting @2MHz, keep it off
	movwf	T1CON
	bcf	PIR1,TMR1IF
	bsf	INTCON,PEIE
	bsf	INTCON,IOCIE
	return

;----------------------- Transmit byte (W):
;(modifies FSR1)
   
UartTxByte
	banksel	0
	movwf	FSR1H			;(save data)
	movf	uartTxWr,W
	addlw	uartTxBuffer
	movwf	FSR1L
	movf	FSR1H,W			;(restore data)
	clrf	FSR1H			;FSR1 => buffer
	movwf	INDF1			;write data to send into buffer

	rlf	INTCON,W		;save GIE -> C
	bcf	INTCON,GIE

	btfsc	T1CON,TMR1ON		;receiver or transmitter busy?
	goto	UartAdjWP		;y: adjust write pointer

	call	UartTxImm		;no, send data immediately, no change of ptr
	goto	UartTxExit

UartAdjWP
	bsf	uartCR,TXBF		;"more data"
	incf	uartTxWr,W		;adjust write pointer
	andlw	UTXBUFSIZE-1
	movwf	uartTxWr

UartTxExit
	btfsc	STATUS,C		;enable interrupt again only if it was enabled
	bsf	INTCON,GIE
	return

;start transmission of byte (INDF1)
;(no change of C-flag)
UartTxImm
	clrf	FSR1H
	movf	INDF1,W
	movwf	uartSR
	bsf	uartCR,TXBSY

	movlw	low (-TMR1RLD+TMR1_LATE2)
	movwf	TMR1L
	movlw	high ((-TMR1RLD+TMR1_LATE2) & 0xFFFF)
	movwf	TMR1H
	bsf	T1CON,TMR1ON

	bsf	uartCR,STASTO		;startbit!
#if (UART_PORT == PORTA)
	banksel IOCAF
	bcf	IOCAN,UART_PIN		;disable Rx
#endif
#ifdef PORTB
 #if (UART_PORT == PORTB)
	banksel IOCBF
	bcf	IOCBN,UART_PIN		;disable Rx
 #endif
#endif

	banksel	LATA
	bcf	UART_PORT,UART_PIN
	banksel	TRISA
	bcf	UART_PORT,UART_PIN
	banksel 0
	return

;----------------------- Interrupt Service:

;TMR1 Interrupt: software UART Rx or Tx
intTMR1
;	movlw	2			;toggle PORTA,1
;	xorwf	PORTA,F			;debug test!

	bcf	PIR1,TMR1IF
	btfsc	uartCR,TXBSY
	goto	intTMR1UartTx		;transmit next bit

	btfsc	uartCR,RXBSY
	goto	intTMR1UartRx

	bcf	T1CON,TMR1ON		;extra time after Rx-stopbit is over
	btfss	uartCR,TXBF		;are there any data to send in buffer?
	retfie				;no

UartTxMore				;y: issue transmission
	movf	uartTxRd,W
	addlw	uartTxBuffer
	movwf	FSR1L
	call	UartTxImm		;transmit next data byte 
	incf	uartTxRd,W
	andlw	UTXBUFSIZE-1
	movwf	uartTxRd		;adjust ptr
	xorwf	uartTxWr,W
	btfsc	STATUS,Z		;buffer empty?
	bcf	uartCR,TXBF		;y: clear flag, no more data!
	retfie

intTMR1UartRx
	bcf	STATUS,C		;read Rx-bit -> C 
	btfsc	UART_PORT,UART_PIN
	bsf	STATUS,C

	btfsc	uartCR,STASTO		;stopbit expected?
	goto	intUartRxStop

	rrf	uartSR,F		;shift data bit into register
	btfsc	STATUS,C		;shifted all 8 bits?
	bsf	uartCR,STASTO		;y: expect stopbit
	goto	intUartReloadT1

intUartRxStop
	movf	uartSR,W
	movwf	uartDR			;received data -> uartDR
	btfsc	STATUS,C
	bsf	uartCR,RXBF		;proper stopbit: data in uartDR seems to be valid
	bcf	uartCR,RXBSY
	call	intUartRxOn		;enable detection of startbit
#ifdef DEBUG_UART
	call	MonRxByte		;process char in monitor program
#endif
	goto	intUartReloadT1		;stop TMR1 after another timeout to force a little break
					;before any pending transmission can start

;UART transmit next bit:
intTMR1UartTx	
	bsf	STATUS,C
	btfsc	uartCR,STASTO
	goto	intUartTxBit		;after startbit: shift a 1-bit into SR

	bcf	STATUS,C		;shift 0-bits into SR
	movf	uartSR,F		;=00 if stopbit was sent
	btfss	STATUS,Z
	goto	intUartTxBit		;not yet stopbit

	bcf	uartCR,TXBSY		;transmission done
	bcf	T1CON,TMR1ON
	btfsc	uartCR,TXBF		;are there any data to send in buffer?
	goto	UartTxMore		;y: transmit

	retfie


intUartTxBit
#ifdef COLLISIONDETECT
	movf	UART_PORT,W		;check actual level of TxD
	banksel	TRISA
	xorwf	UART_PORT,W		;TRIS-bit = set value
	banksel	0
	btfsc	WREG,UART_PIN		
	bsf	uartCR,COLLISION	;set value does not match read value!
#endif
	bcf	uartCR,STASTO
	rrf	uartSR,F

#ifdef SIMULATOR
	banksel	TRISA
	bcf	UART_PORT,UART_PIN	;Simulator: push-pull output
	banksel	PORTA
	btfss	STATUS,C
	bcf	UART_PORT,UART_PIN	;output low
	btfsc	STATUS,C
	bsf	UART_PORT,UART_PIN	;output high
#else
	banksel	LATA
	bcf	UART_PORT,UART_PIN	;emulate open drain output
	banksel	TRISA
	btfss	STATUS,C
	bcf	UART_PORT,UART_PIN	;pull output low
	btfsc	STATUS,C
	bsf	UART_PORT,UART_PIN	;release output (TRIS-bit = 1)
#endif
	banksel	0
	movf	uartSR,W		;just sent stopbit?
	btfsc	STATUS,Z
	call	intUartRxOn		;y: enable Rx startbit detection

intUartReloadT1				;reload timer: interrupt for next bit (Tx or Rx)
#if (TMR1RLD <= 0x100)
	decf	TMR1H,F			
	movlw	TMR1RLD-1
	subwf	TMR1L,F			;8 bits: subtract while timer is running
#else
	movlw	low (TMR1RLD-1)
	bcf	T1CON,TMR1ON		;16bit subtraction is not atomic: must stop timer!
	subwf	TMR1L,F
	movlw	high (TMR1RLD-1)
	subwfb	TMR1H,F
	bsf	T1CON,TMR1ON
#endif
	retfie


intIOC
	btfsc	IOCAF,UART_PIN		;(x = A or B, according to Tx/Rx-Pin location
	goto	IOCUartInt		;H->L transition: UART Rx start bit
	retfie				;(bad interrupt!)

;Interrupt by H->L transition of Rx-Pin: 
IOCUartInt
#if (UART_PORT == PORTA)
	bcf	IOCAF,UART_PIN		;clear flag
	bcf	IOCAN,UART_PIN		;further rx-interrupts by timer...
#endif
#ifdef PORTB
 #if (UART_PORT == PORTB)
	bcf	IOCBF,UART_PIN		;clear flag
	bcf	IOCBN,UART_PIN		;further rx-interrupts by timer...
 #endif
#endif
	banksel	0
	btfss	UART_PORT,UART_PIN
	goto	UartStartRx		;got startbit

;level on rx-pin is high: no startbit here, it was just a glitch!
	call	intUartRxOn		;re-enable detection of H->L
	retfie


UartStartRx
	bcf	T1CON,TMR1ON		;stop (for sure...)
;setup TMR1 interrupt, scan bit0 in the middle of bitframe
	movlw	low (-3*TMR1RLD/2 + TMR1_LATE1)
	movwf	TMR1L
	movlw	high ((-3*TMR1RLD/2 + TMR1_LATE1) & 0xFFFF)
	movwf	TMR1H
	bsf	T1CON,TMR1ON		;set TMR1 runnning
	bcf	PIR1,TMR1IF

	movlw	0x80
	movwf	uartSR
	bsf	uartCR,RXBSY
	bcf	uartCR,TXBSY
	bcf	uartCR,STASTO
	retfie

;---------- sub: enable detection of Startbit (Rx H->L transition)
intUartRxOn
#if (UART_PORT == PORTA)
	banksel	IOCAN
	bcf	IOCAF,UART_PIN
	bsf	IOCAN,UART_PIN		;enable detection of startbit again
#endif
#ifdef PORTB
 #if (UART_PORT == PORTB)
	banksel	IOCBN
	bcf	IOCBF,UART_PIN
	bsf	IOCBN,UART_PIN		;enable detection of startbit again
 #endif
#endif

#ifdef SIMULATOR
	banksel	TRISA
	bsf	UART_PORT,UART_PIN	;sim: set port = input
#endif
	banksel	0
	return

#endif		;(DEBUG_UART)

	end