;**************************************************************
;***      Universal Light Controller:                       ***
;***      Routines for asynchronious serial receiver        ***
;**************************************************************

;The serial communication is used for controlling switches 0..7.
;Data format: 1 startbit, 8 data bits, 1 stopbit
;Protocol:
;Commands start with a '@' character (0x40), followed by an 8 bit
;address, a command code and data for that command. Space chars
;are not allowed. All numerical data are sent in hexadecimal by 
;two ASCII characters per 8 bit value (0..9,A..F upper case only!)
;Commands have to be terminated by a [CR]-character (0x0D). 
;The received address value is compared with the contents of
;"EE_deviceID" (stored in EEPROM), and the controller ignores 
;any command if the address does not match EE_deciceID. This may be
;used for adressing a particular PICALIC module from a number of 
;modules that are all connected to a common serial bus. 
;
;Command "=": set switches (0..7) to new state 
;	@<addr>=<data>
;<data> = new switch data, one bit for each switch, "1" = "on", "0" = "off" 
;Example: @0A=9C will turn switches 2,3,4,7 of module 10 (address 0x0A) to
;the "on" state, while switch 0,1,5,6 are turned "off".

;Command "S": setting combination of switches (0..7) to "on" state 
;	@<addr>S<data>
;<data> = bit mask, each "1" will set the corresponding switch to "on"
;Example: @0AS1C will set switches 2,3 and 4 of module 10 (address 0x0A).

;Command "R": resetting combination of switches (0..7) to "off" state 
;	@<addr>R<data>
;<data> = bit mask, each "1" will reset the corresponding switch to "off"
;Example: @20R03 will reset switches 0 and 1 of module 32 (address: 0x20).

;Command: change module's adress
;	@<addr>PRG<new address>
;<new address> = vale that will be stored in EE_deviceID.
;Normally, after programming the chip, EE_deviceID contains "FF", thus
;command "@FFPRGxx" may be used to initially setup the address of 
;a "new" module (replace xx by your choice...).  
 
#include picalic_config.inc
#include MCU_Headers_em.inc	;Pocessor specific definitions

 if SERIAL_IN

#define BITTIME (1000000/BAUDRATE)
 if CLOCK == 4
  #define LAT1 8		;timing correction/latency values...
  #define LAT2 38
 else	;CLOCK == 8:
  #define LAT1 4
  #define LAT2 19
 endif

#define CR 0x0D			;CR character code

;**************************************************************
;Initialize serial receiver 
;**************************************************************
rx_init
	movlw	low EE_deviceID
	call	ee_read
	movwf	identity

	bsf	STATUS,RP0
	errorlevel-302
	bsf	PIE1,TMR1IE	;interrupt on TMR1 overflow
 ifdef T16F648A
   if SD_INVERT
 	bsf	OPTION_REG,INTEDG ;int on rising edge
   else
	bcf	OPTION_REG,INTEDG ;int on falling edge
   endif
	bcf	STATUS,RP0
   	errorlevel+302
 else
	bcf	STATUS,RP0
   	errorlevel+302
	bcf	RX_PORT,RX_PIN	;(dummy op.) clr mismatch condition 
 endif
	movlw	1
	movwf	cmd_state		;expect command code
	movlw	9
	movwf	rx_bitcnt		;initialize bitcounter (8 data bits)
	movlw 	low(LAT2 - 3*BITTIME/2)	;load delay for scanning first data bit after 
	movwf	TMR1L			;leading edge of start bit
	movlw 	high((LAT2 - 3*BITTIME/2) & 0xFFFF)
	movwf	TMR1H

	bcf	INTCON,RX_INTF
	bsf	INTCON,RX_IEF		;enable interrupt by leading edge of startbit
	return	

;**************************************************************
;RX Interrupt Service Routine: 
;execution times of ISR (instruction cycles):
;start bit transition: 22
;rx data bit: 30|33 (>=4800 Bd | < 4800 Bd)
;rx stop bit + processing data: 45-50 ... 52-57 
;     = (expect char ... process hex char)
;-> assume average = 34/symbol 
;!!! no subroutine calls in order to use only one stack level 
;**************************************************************
serial_isr			
 	btfsc	INTCON,RX_IEF		;is port change int enabled?
	goto	rx_start		;yes: startbit detection

;**************************************************************
;scan input level (triggered by TMR1)

rx_scanbit			;(T=8)
	movwf	save_W3		;context saving: W
	swapf	STATUS,W	;STATUS -> W (no change of flags!)
	movwf	save_SR3	;save STATUS

	bcf	STATUS,RP0	;select bank 0

	bcf	PIR1,TMR1IF	;clr interrupt flag

	decfsz	rx_bitcnt,F
	goto	rx_databit

rx_stopbit
	movlw	9
	movwf	rx_bitcnt

  if SD_INVERT
	btfsc	RX_PORT,RX_PIN	;expect stop bit (low)
  else
	btfss	RX_PORT,RX_PIN	;expect stop bit (high)
  endif
	goto	frerr_ph1	;wrong level: framing error!

	bcf	T1CON,TMR1ON	;stop timer
	movlw 	low(LAT2 - 3*BITTIME/2)	;load delay for scanning first data bit after 
	movwf	TMR1L			;leading edge of start bit
 if(high((LAT2 - 3*BITTIME/2) & 0xFFFF)) == 0xFF
	decf	TMR1H,F
 else
	movlw 	high((LAT2 - 3*BITTIME/2) & 0xFFFF)
	movwf	TMR1H
 endif

	bcf	INTCON,RX_INTF
	bsf	INTCON,RX_IEF		;enable port change int for leading edge of start bit

	movf	cmd_state,W
	incf	cmd_state,F		;default: next state...

;------------------- Jump table for command processing state machine
serialcmdtb
	brw
sertab_def
	goto	frerr_ph2		;0: phase 2: check if all bits were ones (=idle)
	goto	cmdstart		;1: read command startcode
	goto	hexchar			;2: store upper hex-digit value (address)
	goto	hexchar			;3: store lower hex-digit value (address)
	goto	cmdselect		;4: select which command
;(default) Sequence for command "=":
	goto	hexchar			;5: get new switch data (bits 4..7)
	goto	hexchar			;6: ... (bits 0..3)
	goto	cmd_n_cr		;7: <CR> termination of command

;sequence for command "P":
sertab_prg
	goto	check_R			;expect "RG" to complete "PRG"...
	goto	check_G	
	goto	hexchar			;get new device address (bits 4..7)
	goto	hexchar			; ... (bits 0..3)
	goto	cmd_p_cr		;<CR> termination of P cmd

;sequence for command "S":
sertab_s
	goto	hexchar			;get switch data (bits 4..7)
	goto	hexchar			;... (bits 0..3)
	goto	cmd_s_cr		;execute switches OR data -> switches

;sequence for command "R":
sertab_r
	goto	hexchar			;get switch data (bits 4..7)
	goto	hexchar			;... (bits 0..3)
	goto	cmd_r_cr		;execute switches AND ~data -> switches



;*********************************************************************
;processing of framing error (got 0 while stopbit (=1) was expected)
;resumes normal operation after detection of at least 9 consecutive 1-bits:

;processing stopbit while cmd_state = 1: 
frerr_ph2
	incfsz	rx_data,W		;data == 0xFF?
	goto	frerr_rpt		;no: restart sync again
	goto	rxisr_exit2		;yes: assume rx=idle, next byte will be command

frerr_rpt
	bcf	INTCON,RX_IEF		;disable port change int
	bsf	T1CON,TMR1ON		;set timer running
	
frerr_ph1
	clrf	cmd_state

;receive data bit
rx_databit			

 if BAUDRATE >= 4800		;adjust timer "on the fly" on higher baudrates
	movlw 	low(BITTIME-1)
	subwf	TMR1L,F		; reload timer: time for scanning of next bit
	decf	TMR1H,F		;-> 0xFF
	
 else
	bcf	T1CON,TMR1ON	;lower baudrates (> 208 s/bit): stop timer
	movlw 	low(BITTIME-LAT1)
	subwf	TMR1L,F		; reload timer: time for scanning of next bit
	movlw 	high(BITTIME-LAT1)
	btfss	STATUS,C
	addlw	1
	subwf	TMR1H,F
	bsf	T1CON,TMR1ON	;restart timer
 endif
				;(C=0 from last subtract op.) 
  if SD_INVERT
	btfss	RX_PORT,RX_PIN	
  else
	btfsc	RX_PORT,RX_PIN
  endif
	bsf	STATUS,C	; got 1-bit
	rrf	rx_data,F
rxisr_exit2
	swapf	save_SR3,W	
	movwf	STATUS		;restore original STATUS
	swapf	save_W3,F	;(keep Z-flag!)
	swapf	save_W3,W	;restore original W
	retfie

;**************************************************************
;interrupt was triggered by leading edge of startbit

rx_start
	bcf	flags,SAVERP0 
	btfsc	STATUS,RP0	;save RP0-bit...
	bsf	flags,SAVERP0
	bcf	STATUS,RP0	;select bank 0

   if SD_INVERT
	btfss	RX_PORT,RX_PIN	;stable high level?
   else
	btfsc	RX_PORT,RX_PIN	;stable low level?
   endif
	goto	rxisr_exit1	;not stable: ignore glitches
	
;detected leading edge of start bit:
	bsf	T1CON,TMR1ON	;set timer running
	bcf	INTCON,RX_IEF	;disable port change int

rxisr_exit1
	btfsc	flags,SAVERP0
	bsf	STATUS,RP0	;restore RP0
	bcf	INTCON,RX_INTF	;clear interrupt flag
	retfie

;**************************************************************
;processing of just received byte:
;cmd_state selects through "serialcmdtb", which code part has
;to be executed
;**************************************************************

;first byte: expect "@"
cmdstart
	movf	rx_data,W
	xorlw	'@'
	btfss	STATUS,Z
	decf	cmd_state,F	;got some other character: try again with next byte 
	goto	rxisr_exit2

;converts ASCII to hex and assembles byte-value (= 2 hex chars) in rx_value  
hexchar
	swapf	rx_value,W	;move previous nibble to bits 4..7
	andlw	0xF0		;clear lower nibble
	movwf	rx_value

	movf	rx_data,W
	addlw	-0x3A
	btfsc	STATUS,C
	goto	hex_alpha	;char > '9': check for A..F

	addlw	10
	btfss	STATUS,C
	goto	invalid		;char < '0'

	iorwf	rx_value,F	;lower nibble = data	
	goto	rxisr_exit2

hex_alpha
	addlw 	0x3A-0x41
	btfss	STATUS,C
	goto	invalid		;char < 'A'
	addlw	-6
	btfsc	STATUS,C
	goto	invalid		;char > 'F'

	addlw	16		;--> 0x0A..0x0F
	iorwf	rx_value,F	;lower nibble = data
	goto	rxisr_exit2


;---------- evaluate command code: 
cmdselect
	movf	identity,W
	xorwf	rx_value,W	;received right address?
	btfss	STATUS,Z
	goto	invalid		;address does not match -> ignore this command!

	movf	rx_data,W
	xorlw	'='
	btfsc	STATUS,Z
	goto	rxisr_exit2	;command = "=": set new control switch states = data

	movf	rx_data,W
	xorlw	'R'
	btfsc	STATUS,Z
	goto	cmd_r		;command = "R": reset control switches (log. AND with ~data)

	movf	rx_data,W
	xorlw	'S'
	btfsc	STATUS,Z
	goto	cmd_s		;command = "S": reset control switches (log. OR with data)

	movf	rx_data,W
	xorlw	'P'
	btfss	STATUS,Z
	goto	invalid		;got some unsupported cmd

;cmd = "P": set new address
cmd_p	movlw	sertab_prg-sertab_def
	goto	setcmdstate

;check if we got letter "R":
check_R
	movlw	'R'
	goto 	check_char
;check if we got char "G":
check_G
	movlw	'G'
check_char
	xorwf	rx_data,W
	btfsc	STATUS,Z
	goto	rxisr_exit2

;got something wrong: start again scanning for command
invalid	
	movlw 1			;set state = expecting new command 
	goto	setcmdstate

;set execution sequence for cmd "S" 
cmd_s	movlw	sertab_s-sertab_def
	goto	setcmdstate

;set execution sequence for cmd "R" 
cmd_r	movlw	sertab_r-sertab_def

setcmdstate
	movwf	cmd_state
	goto	rxisr_exit2

;expect proper termination of "P" command and execute 
cmd_p_cr
	movf	rx_data,W
	xorlw	CR
	btfss	STATUS,Z
	goto	invalid		;did not get right termination character

;----------------- write new device ID into EEPROM 
	bsf	STATUS,RP0		;select bank 1
 errorlevel -302
	movlw	low EE_deviceID
	movwf	EEADR
		
	movf	rx_value,W
	movwf	EEDATA			;= new data
	movwf	identity		;adjust value in RAM

	bsf	EECON1,WREN		;disable write protection
	movlw	0x55
	movwf	EECON2			;issue write op...
	movlw	0xAA
	movwf	EECON2
	bsf	EECON1,WR
	bcf	EECON1,WREN		;enable write protection again
ee_wait4wr
	btfsc	EECON1,WR
	goto	ee_wait4wr		;wait for write-op complete

	bcf	STATUS,RP0		;-> bank 0
 errorlevel +302
	
;show LEDs flashing for confirmation of successful programming:
	movlw 	32
	movwf	rx_data			;loop counter for outer loop

cmd_p_flashloop
 if ACTIVELOW
	movlw	0xFF
 else
	movlw	0			;LEDs = off
 endif	
	btfss	rx_data,1
	goto	cmd_p_flash0
	btfsc	rx_data,2
	
	xorlw	0xFF			;turn LEDs on
cmd_p_flash0
	movwf	CH0_PORT
	
cmd_p_flash1				;inner loop:
	btfss	INTCON,T0IF		;wait fot TMR0 timeout
	goto	cmd_p_flash1
	bcf	INTCON,T0IF
	incfsz	rx_value,F
	goto	cmd_p_flash1		;delay = 64 or 32 ms (4 or 8 MHz clock)

	decfsz	rx_data,F		;next outer loop
	goto	cmd_p_flashloop

 	goto	invalid			;reset cmd status -> expect new command


;expect proper termination of "N" command and process data 
cmd_n_cr
	movf	rx_data,W
	xorlw	CR
	btfss	STATUS,Z
	goto	invalid			;did not get right termination character

	movf	rx_value,W
	movwf	switch
	goto	invalid			;reset cmd status -> expect new command

;expect proper termination of "S" command and process data
cmd_s_cr
	movf	rx_data,W
	xorlw	CR
	btfss	STATUS,Z
	goto	invalid			;did not get right termination character

	movf	rx_value,W
	iorwf	switch,F
	goto	invalid			;reset cmd status -> expect new command

;expect proper termination of "R" command and process data
cmd_r_cr
	movf	rx_data,W
	xorlw	CR
	btfss	STATUS,Z
	goto	invalid			;did not get right termination character

	comf	rx_value,W
	andwf	switch,F
	goto	invalid			;reset cmd status -> expect new command
	
 endif		;SERIAL_IN

	END