;***************************************************************************
; Serial Debug Monitor
;***************************************************************************
;commands (not case sensitive):
;R<addr> [<addr> ...]			Read contents of memory or SFR
;W<addr> <data> [<addr> <data> ...]	Write data into memory or SFR
;P					Pause (stops main program)
;G					Go (resumes execution of main program)

;Input: from UART Rx ISR, this routine must call MonRxByte
;Output: calls UartTxByte in order to send byte (W) to host.

  ifdef __16F1825
	#include <P16F1825.INC>         	; processor specific variable definitions
  endif
#include picalic_config.inc

#ifdef DEBUG_UART

	extern uartCR, uartDR, UartTxByte
	global MonRxByte

	radix dec

;uartCR status Bits:
#define RXBSY 0		;receiving
#define TXBSY 1		;transmitting
#define RXBF 2			;receiver buffer full (rx-data in uartDR)
#define TXBF 3			;txbuffer contains data to send
#define STASTO 4		;rx: expect stopbit, tx: just sent startbit 
#define COLLISION 5		;set to 1 if a data collision was detected (only

MONVARS	udata 0x40		
;Variables in bank 0, initialize to 0 upon RESET
mCommand res 1		;command code
mState	res 1		;actual command state (0: idle), upper bits: flags
mVal 	res 2
mAddr	res 2
mInstrAddr res 2
mTemp res 1				


;mState flag bits:
MGOTVAL	equ	7		;some value was entered

MONI	code
 errorlevel-302

;====================================================
;Entry point - called with received data byte in uartDR:
 
MonRxByte
	bcf	uartCR,RXBF
	movf	uartDR,W
	addlw	-0x60		
	btfsc	STATUS,C	;lower case letter?
	bcf	uartDR,5	;y: convert to upper case

;execute code according to mState:
MonJmpState
	movf	mState,W
	andlw	0x1F		;upper bits are flags
	brw
	goto	MonDecodeCmd	;0: state=idle, first char is command
	goto	MonGetVal	;1: read hex number -> mVal
	goto	MonRead		;2: read register (mVal) 
	goto	MonGetVal	;3: Write register: get address
	goto	MonWrStAdr	;4: Write register: store address
	goto	MonGetVal	;5: Write register: get data to write
	goto	MonWrExecute	;6: Write register: do the write op.	

MonDecodeCmd
	call	clr_mVal
	movlw	N_COMMANDS+1
	movwf	mCommand
MonCmdDecodeLp
	decfsz	mCommand,F	;=0 if cmd is invalid
	goto	MonCmdDec1
MonInvalid
	movlw	'?'
	call	UartTxByte
	clrf	mState		;-> idle
	movlw	0x0A
	goto	UartTxByte

MonCmdDec1		
	call	MonCmdChar	;W = MonCmdList[mCommand-1]
	xorwf	uartDR,W	;found matching cmd?
	btfss	STATUS,Z	
	goto	MonCmdDecodeLp	;no, try next from list

	call	MonDecodeExec
	movwf	mState
	return		

MonDecodeExec
	decf	mCommand,W
	brw
	goto	MonExecGo	;'G'
	goto	MonExecPause	;'P'
	retlw	0		; LF: ->idle
	retlw	3		;'W':get address...
	retlw	1		;'R':get address to read


;returns command character for code (mCommand)
MonCmdChar
	decf	mCommand,W
	brw
MonCmdList
	retlw	'G'		;Go: resume program execution
	retlw	'P'		;Pause: stop program execution
	retlw	0x0A		;line feed: idle mode without '?'-message
	retlw	'W'		;set memory or SFR to <data>
	retlw	'R'		;read memory or SFR
N_COMMANDS equ $-MonCmdList


;------------------------------------------------------------------------
;process chracter for entering hexadecimal values.
;calls next "mState" routine if number has been entered sucessfully
MonGetVal
	call	ishex
	btfss	STATUS,C
	goto	MonNotHexCh	;this char is no hex char

	bsf	mState,MGOTVAL
	movwf	uartDR		;= 0..0x0F
	swapf	uartDR,F
	movlw	4
MonGHLp
	rlf	uartDR,F
	rlf	mVal,F
	rlf	mVal+1,F
	decfsz	WREG,W
	goto	MonGHLp
	return	

MonNotHexCh
	incf	mState,F	;default: do the next step
	btfsc	mState,MGOTVAL	;was some number entered?
	goto 	MonJmpState	;y: process entered number...

;got no hex character while value is still empty:
	decf	mState,F
	movlw	' '
	xorwf	uartDR,W
	btfsc	STATUS,Z
	return			;ignore leading space chars
	
	clrf	mState		;-> idle
	goto	MonRxByte	;try to interpret as new command

;---------------------------------------------------
;send contents of RAM or SFR
MonRead
	decf	mState,F	;continue "Read" cmd.
	movf	mVal,W
	movwf	FSR0L
	movf	mVal+1,W
	movwf	FSR0H
	call	clr_mVal
	
MonSpcHexOut
	movlw	' '
	call	UartTxByte
MonHexOut
	swapf	INDF0,W
	call	MHexOut1
	movf	INDF0,W
MHexOut1
	andlw	0x0F
	addlw	-10
	addlw	10
	iorlw	0x30
	btfss	STATUS,C
	addlw	'A'-0x3A
	goto	UartTxByte		


;---------------------------------------------------
;write reg, part 1: store address
MonWrStAdr
	movf	mVal,W
	movwf	mAddr
	movf	mVal+1,W
	movwf	mAddr+1
	call	clr_mVal
	incf	mState,F	;next step: enter data value
	return

;---------------------------------------------------
;write reg, part 2: write data to register
MonWrExecute
	movf	mAddr,W
	movwf	FSR0L
	movf	mAddr+1,W
	movwf	FSR0H
	movf	mVal,W
	movwf	INDF0
	movlw	3
	movwf	mState		;repeat (W addr data addr data...)
clr_mVal
	clrf	mVal
	clrf	mVal+1
	bcf	mState,MGOTVAL
	return

;---------------------------------------
;check if usrtDR is a hexadecimal char (0..9,A..F)
;if yes: returns C=1, value (0..0x0F) in W
ishex
	movlw	'0'
	subwf	uartDR,W
	btfss	STATUS,C
	return			;< '0': no hex char, C=0

	addlw	-10
	addlw	10
	btfsc	STATUS,C
	return			;0..9

	addlw	-('A'-'0')
	btfss	STATUS,C
	return			;<'A'
;if char = 'A'..'F', W contains 0..5 here:
	addlw	-6		;negative for valid hex chars
	addlw	6+10		;adjust to 0x0A..0x0F, or C=0 if char was >'F'
	return
	

;---------------------------------------------------
;Break: stop execution of main program
;monitor is called from UART Rx-Interrupt. Main prg. is stopped by replacing return address on stack
;with address of "goto $" instruction 
MonExecPause
	movlw	mInstrAddr+1
	movwf	FSR0L
	clrf	FSR0H

	movf	mInstrAddr,W
	iorwf	mInstrAddr+1,W		;already in brk state?
	btfss	STATUS,Z
	goto	MonBrkSkip		;y: skip

	banksel	TOSL
	movlw	-2
	addwf	STKPTR,F		;STKPTR -> ISR return address
	movf	TOSL,W			;save return address
	movwi	-1[FSR0]		
	movf	TOSH,W			
	movwi	0[FSR0]
	movlw	low MonDummy		;change return address to dummy
	movwf	TOSL
	movlw	high MonDummy
	movwf	TOSH
	movlw	2
	addwf	STKPTR,F		;restore STKPTR

MonBrkSkip
	call	MonHexOut
	decf	FSR0L,F
	call	MonHexOut		
	retlw	0			;next state = idle

MonDummy
	goto	$

;---------------------------------------------------
;Go: resume execution of main program
MonExecGo
	movlw	mInstrAddr+1
	movwf	FSR0L
	clrf	FSR0H

	movf	mInstrAddr,W
	iorwf	mInstrAddr+1,W		;in brk state?
	btfsc	STATUS,Z
	retlw	0			;n: ignore, -> idle

	banksel	TOSL
	movlw	-2
	addwf	STKPTR,F		;STKPTR -> ISR return address
	moviw	-1[FSR0]
	movwf	TOSL			;restore return address
	moviw	0[FSR0]
	movwf	TOSH			
	movlw	2
	addwf	STKPTR,F		;restore STKPTR
	banksel	mInstrAddr
	clrf	mInstrAddr
	clrf	mInstrAddr+1
	retlw	0			;next state = idle

#endif	; DEBUG_UART

	end
	